// Programm:   Rat_F.cpp zur Abschtzung des Auswertefehlers einer
// rationalen Funktion mit einem Faktor bei hochgenauer Arithmetik. 
#include "bnd_util.hpp"   // Wegen Rat_F(), Poly_Dat_rel(), ...
#include <iostream>       // Wegen cout

using namespace cxsc;
using namespace std;

real Rat_F_Xi(const interval& Xi, const real& epsx, 
              const ivector& ia, const ivector& ib, 
              const int& k, const real& x0)
// x * P_N(u)/Q_M(u);  u = (x-x0)^k,  1<=k<=6;
// Berechnung der relativen Fehlerschranke ber einem Teilintervall Xi,
// welches die exakten Polynom-Argumente enthlt.
// Vorausgesetzt wird eine nur hochgenauer Arithmetik, mit u(x) = (x-x0)^k. 
// epsx ist die relative Fehlerschranke der gestrten Polynomargumente.
// ia,ib sind Punktintervalle und importieren die optimalen Einschlieungen
// der exakten Polynomkoeffizienten von P_N(u) bzw. Q_M(u). 

{
    real del1,del2,delu,delK;
    interval Kl,u,h1,h2;
    abs_subh3(Xi,epsx,interval(x0),0,Kl,delK); 
    // Kl  : Einschlieung von (x-x0) fr alle x aus Xi;
    // delK: Absolute Fehlerschranke bei Auswertung von x-x0, wobei die 
    //       relative Fehlerschranke epsx der gestrten Polynomargumente 
    //       bercksichtigt wird. 
    // Auswertung von u(x) = (x-x0)^k:
    switch (k)
    {
	case 1:  // u(x) = (x-x0)^1
	    u = Kl;  delu = delK;
	    break;
	case 2:  // u(x) = (x-x0)^2
	    abs_mulh1(Kl,delK,Kl,delK,u,delu);
	    break;
	case 3:
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1);  // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,Kl,delK,u,delu);   // u: Einschg. fr (Kl*Kl)*Kl
	    break;
	case 4:  // u(x) = x^4
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1);  // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,u,delu);  // u:  Einschlg. fr (Kl*Kl)^2
	    break;
	case 5:  // u(x) = x^5
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1); // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Kl*Kl)^2
            abs_mulh1(h2,del2,Kl,delK,u,delu);  // u:  Einschlg. fr Kl^5
	    break;
	case 6:  // u(x) = x^6
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1); // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Kl*Kl)^2
            abs_mulh1(h2,del2,h1,del1,u,delu);  // u:  Einschlg. fr Kl^6
	    break;
	default:
	    cerr << "Horn_relh_Xi(): Unzulssige Potenz x^k" << endl;
	    exit(1);
    } // switch
    // u ist Einschlieung von (Xi-x0)^k; 
    // delu ist die zugehrige absolute Fehlerschranke; 

    h1 = ia[Ub(ia)]; // h1: Horner-Startwert fr das Zhlerpolynom
    del1 = 0;  // Die Polynomkoeffizienten sind rundungsfehlerfrei!
    int m = Ub(ia)-1;
    for (int j=m; j>=0; j--)
    {   
	abs_mulh1(h1,del1,u,delu,h2,del2);          // h2 = y*u
	if (ia[j]==1) abs_1px_delh(h2,del2,h1,del1);
	else abs_addh1(ia[j],0,h2,del2,h1,del1); // h1 = y*u + a[j]
    } 
//********************************************************************
// -------- Bercksichtigung des Faktors x im Zhler durch: ----------
    abs_mulh3(Xi,epsx,h1,del1,Kl,delK);
//     Kl  : Einschlieung von Xi*Zhlerpolynomwert
//     delK: Absolute Fehlerschranke des Zhlers  Xi*P_N(u)
//********************************************************************

    h1 = ib[Ub(ib)]; // h1: Horner-Startwert fr Nennerpolynoms
    del1 = 0;  // Die Polynomkoeffizienten sind rundungsfehlerfrei!
    m = Ub(ib)-1;
    for (int j=m; j>=0; j--)
    {   
	abs_mulh1(h1,del1,u,delu,h2,del2);          // h2 = y*u
	if (ib[j]==1) abs_1px_delh(h2,del2,h1,del1);
	else abs_addh1(ib[j],0,h2,del2,h1,del1); // h1 = y*u + a[j]
    } // del1: absoluter Fehler;  h1: Einschlieung des Nennerpolynoms

    // Berechnung des relativen Fehlers des Polynom-Quotienten:
    rel_divh1(Kl,delK,h1,del1,h2,del2);

    return del2;  // Rckgabe der relativen Fehlerschranke del2
} // Rat_F_Xi

real Rat_F(const rvector& a, const rvector& b, const real& x0, 
           const interval& X, const real& epsx, const int& k, 
           const real& diam)
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn der
// Quotient    P_N(u)/Q_M(u),  u=(x-x0)^k,  1<=k<=6, mit den
// Polynomen:  P_N(u) = a[0] + a[1]*u^1 +...+a[N]*u^N,  
//             Q_M(u) = b[0] + b[1]*u^1 +...+b[M]*u^M,
// zustzlich noch mit einem Faktor, z.B. x, multipliziert und die 
// Polynome nach dem Hornerschema ausgewertet werden. 
// Wie dieser Faktor bei der Fehlerabschtzung zu bercksichtigen
// ist, zeigt die Implementierung der Funktion Rat_F_Xi() im Quelltext
// des Programms Rat_F.cpp
// a ist der Vektor der exakten Koeffizienten a[j], j=0,1,...,N;
// b ist der Vektor der exakten Koeffizienten b[j], j=0,1,...,M;
// x0 ist der Polynom-Entwicklungspunkt; x0 in S(2,53) muss erfllt sein!
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.
{
    real bnd;
    ivector ia(0,Ub(a)), ib(0,Ub(b));  // Einschlieungen der exakten 
                       // Polynom-Koeffizienten durch Punktintervalle
    for (int j=0; j<=Ub(a); j++) ia[j] = interval(a[j]);
    for (int j=0; j<=Ub(b); j++) ib[j] = interval(b[j]);
    Max_Rat_Xi(Rat_F_Xi,X,epsx,diam,ia,ib,k,x0,bnd);
    return bnd;
} // Rat_F

int main()
{
// Gesucht: Relativer Auswertefehler einer Funktion 
//          R(u) = x*P_N(u)/Q_M(u),  u = (x-x0)^k,  1<=k<=6;
// P_N(u) = a[0] + a[1]*u^1 + ... + a[N]*u^N,  u = (x-x0)^k;
// Q_M(u) = b[0] + b[1]*u^1 + ... + b[M]*u^M,  u = (x-x0)^k;
// a[j],b[j]: Exakte Koeffizienten, die durch Rundung von 
//            Dezimalwerten zur jeweils nchsten Rasterzahl des
//            IEEE-double-Formats entstehen.
// Der Anwender muss die Werte von  N,M,k,x0  hier eingeben: 
    int N=4, M=4, k=2;  // N,M : Polynomgrad; 
                        // u(x)=(x-x0)^k, mit 1<=k<=6;
    real x0 = 0;   // x0: Entwicklungspunkt muss Maschinenzahl sein!
    //string strPj; // N+1 dezimale Komponenten von P_N(u)
    // Der Anwender muss die dezimalen Nherungen der Koeffizienten  
    // von P_N(u) hier eingeben:  z.B. strP0 = "1.01e-3";
    string strP0 = "1.12837916709551256e+0";
    string strP1 = "1.35894887627277916e-1";
    string strP2 = "4.03259488531795274e-2";
    string strP3 = "1.20339380863079457e-3";
    string strP4 = "6.49254556481904354e-5";

    //string strQj; // M+1 dezimale Komponenten von Q_M(u)
    // Der Anwender muss die dezimalen Nherungen der Koeffizienten
    // von Q_M(u) hier eingeben:  z.B. strQ0 = "-1.22301";
    string strQ0 = "1.00000000000000000";
    string strQ1 = "4.53767041780002545e-1";
    string strQ2 = "8.69936222615385890e-2";
    string strQ3 = "8.49717371168693357e-3";
    string strQ4 = "3.64915280629351082e-4";

    interval X;  // Intervall der exakten Polynom-Argumente x
    real bnd, diam, epsx;
    rvector a(0,N),b(0,M); // a[j],b[j]: Exakte Koeffztn. aus S(2,53);

    strP0 >> RndNext >> a[0];
    strP1 >> RndNext >> a[1];
    strP2 >> RndNext >> a[2];
    strP3 >> RndNext >> a[3];
    strP4 >> RndNext >> a[4];

    strQ0 >> RndNext >> b[0];
    strQ1 >> RndNext >> b[1];
    strQ2 >> RndNext >> b[2];
    strQ3 >> RndNext >> b[3];
    strQ4 >> RndNext >> b[4];
    // Die a[j],b[j] sind die zur jeweils nchsten Rasterzahl
    // gerundeten Strings strPj,strQj,  j = 0,1,...,4;

    while(1) // Endlosschleife
    {
        Poly_Dat_rel(X,epsx,diam);  // Eingabe von X,epsx,diam;
	    // X   : Einschlieung der exakten Polynom-Argumente
            // epsx: relative Fehlerschranke der gestrten Argumente
            // diam: Mantissendurchmesser der Teilintervalle (10^(-5))
        bnd = Rat_F(a,b,x0,X,epsx,k,diam); // bnd: relative Fehler-
        // schranke bei nur hochgenauer Arithmetik beim Hornerschema
        cout << RndUp
             << "Relat. Fehlerschranke bei hochgenauer Arithmetik = "
	     << bnd << endl << endl;
    }
}


















