// Programm:  book_expl26.cpp
#include <iostream>   // For cout, cin;
#include <imath.hpp>  // For lnp1(A)

using namespace cxsc;
using namespace std;

static const real eps_lnp1 = 
  5087233976787912.0 / 20282409603651670423947251286016.0; 
  // relative error bound of the function ln(1+x); 2.5082E-016;

static const interval MR_l = interval(succ(-MinReal),MinReal); 

static real lnp1_abs4Max(const real& a, const real& dela)
    // For calculating the maximum; used by abs_lnp1_abs()
{
    interval z;
    real r,r1;
    r1 = addd(1,a);
    r = subd(r1,dela);
    r = divu(dela,r);
    z = abs( lnp1(interval(a)) );
    r1 = addu(Sup(z),r);
    r1 = mulu(r1,eps_lnp1);
    r1 = addu(r1,r);
    return r1;
}

void abs_lnp1_abs(const interval& A, const real& dela,
                 interval& R, real& abs)
{
    interval At,D,z=interval(-dela,dela); 
    At = A+z;  D = At+z;  // At = A_tilde
    if (Inf(D)<=-1) 
    {
        cerr << "abs_lnp1_abs(): Absolute error doesn't exist!" 
             << endl;   exit(1);
    }
    real r,r1;
    if (Disjoint(At,MR_l)) // intersection is empty:
    { 
        abs = lnp1_abs4Max(Inf(A),dela);
        r   = lnp1_abs4Max(Sup(A),dela);
        if (r>abs) abs = r;
    }
    else // intersection not empty:
    {
    // Calculating an absolute upper bound abs only for the exact 
    // function values ln(1+wa) lying in the denormalized range:
        D = At & MR_l;  // D: not empty intersection of At and MR_l;
        r = subd(Inf(D),2*dela);
        r = addd(1,r);  // r>0 is guaranteed by: if (Inf(D)<=-1) { }
        r = divu(dela,r);
        abs = addu(r,minreal);  // abs: the desired error bound

        if (Inf(At) <= -MinReal)
        {
            r  = lnp1_abs4Max(Inf(A),dela);
            r1 = addu(-MinReal,dela);
            r1 = lnp1_abs4Max(r1,dela);
            if (r>r1) r1 = r;
            if (r1>abs) abs = r1;
        }

        if (Sup(At)>MinReal)
        {
            r  = lnp1_abs4Max(Sup(A),dela);
            r1 = subd(succ(MinReal),dela);
            r1 = lnp1_abs4Max(r1,dela);
            if (r>r1) r1 = r;
            if (r1>abs) abs = r1;
        }
    } 
    R = lnp1(A); // R: inclusion of ln(1+A);
}
int main()
{
    real abs,dela;
    interval A,R;
    while(1) // infinite loop
    {
        cout << "Intervall A = ? ";  cin >> A;
        cout << "Delta(a)  = ? "; cin >> RndUp >> dela;
        abs_lnp1_abs(A,dela,R,abs);
        cout << "Absolute error = " << RndUp << abs << endl << endl;
    }
}
