// Programm coth.cpp zur Auswertung der Funktion
// coth(x) fr Intervallargumente vom Typ l_interval

#include <imath.hpp>
#include <l_imath.hpp>
#include <iostream>

using namespace std;
using namespace cxsc;

// Fields for coth-function
int  ex_coth[8] = { -1,-5,-8,-12,-15,-18,-22,-25 };  
int  c_cothN[8] = { 1,-1,2,-1,2,-1382,4,-3617 };
real c_cothD[8] = { 1,15,315,1575,31185,212837625,  // all components
                    6081075,54273594375.0 };        // exactly stored!

l_interval Coth(const l_interval & x) throw()
// Coth(x); Blomquist 17.04.04;           
{
    l_interval t,s,c,y;
    interval dx = _interval(x),
             einfachgenau,r,r2;
    int stagsave = stagprec,ex,m,N,k,
        stagmax = 19;
    bool neg;
    einfachgenau = coth(dx); // produces all necessary error messages!

    if (stagprec == 1) y = coth(dx);
    else 
    {
        if (stagprec>stagmax) stagprec = stagmax;
        neg = Sup(dx)<0.0;
        t = x;
        if (neg) { t = -t; dx = -dx; } // Inf(t),Inf(dx) > 0;
        ex = expo(Inf(dx));
        if (ex<-66) { // Laurent series if Inf(dx)<2^(-67)
            m = -ex - 53*stagprec;
            r = interval(Sup(dx)); r2 = r*r;
            N = 0;
            do { 
                N++;
                if (N>7) { y = einfachgenau; goto Fertig; }
                r = r*r2;
                k = expo(Sup(r)) + ex_coth[N];
            } while (k>m);
            r2 = interval(c_cothN[N])/c_cothD[N]; 
            r2 = r*abs(r2)/3;  // r2: inclusion of the absolute error
            N--; // Polynomial degree
            y = l_interval(c_cothN[N])/c_cothD[N]; 
            s = t*t;
            for (k=N-1; k>=0; k--)
                y = y*s + l_interval(c_cothN[k])/c_cothD[k];
            y = (y*t)/3 + 1/t;
            y = y + interval(-Sup(r2),Sup(r2)); // with approx. error
        } else if (ex < 2) {
            if (stagprec<stagmax) stagprec++; // stagprec <= 19
            c = cosh(t);
            s = sinh(t);
            y = c/s;
        } else if (Inf(dx)<353.0) {
            if (stagprec<stagmax) stagprec++; // stagprec <= 19
            times2pown(t,1);
            y = 1.0 + 2.0/(exp(t)-1.0);
        } else { // Inf(dx) >= 353.0
            y = l_interval(1.0) + comp(0.5,-1016);
            SetInf(y,1.0);
        }
        if (_interval(1.0) <= y)  SetInf(y,1.0);
        if (neg) y = -y;
    Fertig:     stagprec = stagsave; // restore old stagprec value
        y = adjust(y);  // adjust precision of y to old stagprec value
        y = y & einfachgenau;
    }
    return y;
} // coth

int main()
{
    l_interval x,y;
    real r1,r2;
    stagprec = 2;  // stagmax = 19

    while (1) {
        cout << "real r1 = ?" << endl;   cin >> r1;
        cout << "real r2 = ?" << endl;   cin >> r2;
        x = l_interval(r1,r2);  // x = l_interval(21)/10;
        y = Coth(x);
        cout << SetDotPrecision(16*stagprec,16*stagprec) 
             << Scientific << "Coth(x) = " << y << endl << endl;
    };
}
