// Programm:  exp_x2_bnd.cpp;
// Berechnung der rel. Fehlerschranke eps(f) in 
// 0 <= x <= x_0 = 26.61571750...  fuer e^(-x^2);
#include "abs_relh.hpp" // Wegen abs_mulh1(),...
#include "bnd_util.hpp" // Wegen Max_bnd_Xi()
#include "hilfe.hpp"    // Wegen sqr2uv(...)
#include <imath.hpp>    // Wegen exp(...)
#include <iostream>     // Wegen cout

using namespace cxsc;
using namespace std;

real T_xA1(const interval& A1)
{
    interval z;
    z = interval(Sup(A1));
    z = z*z;
    z = z*exp(z);
    return Sup(z);
}

real T_xA2(const interval& xi, const real& delx)
{   // e^(-x^2) = 1 - x^2*(1-x^2/2*[1-x^2/3])
    real delu,delz1,delz2,rel_g2,r;
    interval u,z1,z2;
    // Berechnung des Auswertefehlers von g2
    // nach dem Hornerschema:
    abs_mulh1(xi,0,xi,0,u,delu);       // u: Einschliessung von xi*xi
    abs_divh1(u,delu,interval(3),0,z1,delz1);           // z1: x^2/3;
    abs_1mx_delh(z1,delz1,z2,delz2);                  // z2: 1-x^2/3;
    abs_mulh1(u,delu,z2,delz2,z1,delz1);        // z1: x^2*(1-x^2/3);
    times2pown(z1,-1);  times2pown(delz1,-1); // z1: x^2/2*(1-x^2/3);
    abs_1mx_delh(z1,delz1,z2,delz2);      // z2: 1 - x^2/2*(1-x^2/3);
    abs_mulh1(u,delu,z2,delz2,z1,delz1); 
                                      // z1: x^2*[1-x^2/2*(1-x^2/3)];
    rel_1mx_delh(z1,delz1,z2,rel_g2); 
    // rel_g2: Rel. Auswertefehler von g_2(x)
    // Berechnung des Approximationsfehlers:
    z2 = interval(Sup(xi));  z2 = z2*z2;
    z2 = interval(Sup(z2));  // z2: x0^2
    z1 = z2*z2;  z1 = z1*z1; // z1: x0^8
    z1 = z1*exp(z2)/24;  // z1: Relativer Approximationsfehler;
    // Berechnung des relativen Gesamtfehlers:
    z2 = interval(rel_g2);
    z1 = z1 + z2*(1+z1);

    return Sup(z1);  // Rueckgabe des relativen Gesamtfehlers;
}

real T_xA3(const interval& xi, const real& delx)
{   // xi: Teilintervall von X; "e^(-u) + e^(-u)*(-v+delta)"
    // Mit Skalierung!
    const real c1 = 2.35796256e-16; // rel. Fehler von exp(...)
    interval xi2,exp_xi2,vi,z,S;
    real delh,dels,rel,absF,Sxi2,V,r;
    xi2 = xi*xi;  Sxi2 = Sup(xi2);
    exp_xi2 = exp(-xi2);          // e^(-u) in exp_xi2 enthalten
    times2pown(exp_xi2,1020);     // Skalierung mit 2^(1020)
    absF = mulu(c1,Sup(exp_xi2)); // absF: Abs. Fehler bez. e^(-u);
    V = subu(Sxi2,pred(Sxi2));
    vi = interval(-V,V);          // alle v in vi enthalten
    z = interval(V);
    z = (z*z/2)/(1-z);  r = Sup(z);  // r: abs. Approximationsfehler;
    vi = vi + interval(-r,r);        // vi: Einschliessung von:
                                     //     e^(-v)-1 = -v + delta
    abs_mulh1(exp_xi2,absF,vi,r,S,dels);    // S: e^(-u)*(-v+delta);
    rel_subh1(exp_xi2,absF,S,dels,z,rel);

    return rel;
}

int main()
{ 
    const real x0 = 7491658466053896.0 / 281474976710656.0;
    // x0 approx 26.61571750925125;  Die berechnete Fehlerschranke
    // e(f) = 4.618919E-16  gilt fuer alle Maschinenzahlen x <= x0
    interval 
	A1 = interval(0,comp(0.5,-25)), 
	A2 = interval(comp(0.5,-25),comp(0.5,-5)),
	A3 = interval(comp(0.5,-5),x0); 
    real bnd,bnda, diam = 1e-5, delx=0;

    bnd = T_xA1(A1);
    cout << "Bereich A1: Rel. Schranke eps(f,A1) = " 
         << RndUp << bnd << endl;

    Max_bnd_Xi(T_xA2,A2,delx,diam,bnda);
    cout << "Bereich A2: Rel. Schranke eps(f,A2) = " << bnda << endl;
    if (bnda>bnd) bnd = bnda;

    Max_bnd_Xi(T_xA3,A3,delx,diam,bnda);
    cout << "Bereich A3: Rel. Schranke eps(f,A3) = " << bnda << endl;
    if (bnda>bnd) bnd = bnda;

    cout << "eps(f) = " << bnd << endl;
}
