// Programm:  sqrtx2m1_A3.cpp
// Berechnung des relativen Gesamtfehlers
// y0 +0.5*(x-y0^2)/y0
// x = x1+x2;   y0 = s1+s2;
// x-y0^2 = [(x1^2-s1^2)-1] + [x2*(x+x1)-s2*(y0+s1)]
#include "abs_relh.hpp" // Wegen abs_divh1()
#include "bnd_util.hpp" // Wegen Max_bnd_Xi()
#include "hilfe.hpp"    // Wegen Cut26()
#include <iostream>     // Wegen cout

using namespace cxsc;
using namespace std;

real T_x(const interval& Xi, const real& delx)
{
    interval Y0,B,C,D,E,X1,X2,S1,S2,G,H,P,S;
    real r1,r2,rel,app,delb,delc,delh,dele,delg,dels,ep0;
    int ex;

    r1 = Cut26(Inf(Xi));  r2 = Cut26(Sup(Xi));
    X1 = interval(r1,r2); // x = x1+x2;  x1=Cut26(x) enthalten in X1;
    ex = expo(Sup(Xi));
    r1 = comp(0.5,ex-25);
    X2 = interval(0,r1);  // x1 in X1 und x2 in X2 enthalten!

    H = X1*X1; // Fr alle x1 aus X1 wird x1*x1 exakt berechnet!!
    abs_subh1(H,0.0,interval(1),0.0,G,delg); // G = X1*X1-1;  delg=0
    abs_addh1(Xi,0.0,X1,0.0,E,dele);  // E = x + x1
    abs_mulh1(X2,0.0,E,dele,S,dels);  // S = x2*(x + x1)
    abs_addh1(G,delg,S,dels,C,delc);  // C = (x1^2-1)+x2*(x+x1)=x^2-1
    rel_sqrt_abs(C,delc,Y0,ep0);   // Startwert y0=sqrt(x^2-1) in Y0;
    // ep0: relative Fehlerschranke fr den Startwert.
    D = interval(-ep0,+ep0);
    D = 0.5 * D*D/(1+D);
    app = Sup(D);  // app: Relativer Approximationsfehler bez. y1

    r1 = Cut26(Inf(Y0));  r2 = Cut26(Sup(Y0));
    S1 = interval(r1,r2); // y0 = s1+s2;  s1 = Cut26(y0) aus S1;
    ex = expo(Sup(S1));
    r1 = comp(0.5,ex-25);
    S2 = interval(0,r1);  // s1 in S1 und s2 in S2 enthalten!!

    B = S1*S1; // s1*s1 ist rundungsfehlerfrei, denn 2*26=52 < 53
    abs_subh1(H,0.0,B,0.0,P,r2);      // P = x1^2 - s1^2;
    abs_subh1(P,r2,interval(1),0.0,D,r1); // D = (x1^2 - s1^2) - 1;
    abs_addh1(Y0,0.0,S1,0.0,B,delb);  // B = y0 + s1;
    abs_mulh1(S2,0.0,B,delb,G,delg);  // G = s2*(y0 + s1);
    abs_subh1(S,dels,G,delg,B,delb);  // B = x2*(x + x1)-s2*(y0 + s1)
    abs_addh1(D,r1,B,delb,S,dels);    // S = x - y0^2
    abs_divh1(S,dels,Y0,0.0,C,delc);  // C = (x-y0^2)/y0
    C = C/2;  delc = delc/2;
    rel_addh1(Y0,0.0,C,delc,B,rel);

    // Bercksichtigung des Approximationsfehlers:
    r1 = addu(1.0,app);
    r2 = mulu(rel,r1);
    rel = addu(app,r2);

    return rel; // rel: Schranke des relativen Gesamtfehlers.
}

int main()
{
    interval X = interval(1024,44000);
    real bnd, diam = 1e-6, delx=0;  
    Max_bnd_Xi(T_x,X,delx,diam,bnd);
    cout << RndUp << "Relative Fehlerschranke = " << bnd << endl;
}
